import numpy as np
from numpy.random import beta
import sys
#sys.path.append('../h5hep')
#from write import *
import hepfile

################################################################################
def calc_energy(mass,px,py,pz):

    energy = np.sqrt(mass*mass + px*px + py*py + pz*pz)

    return energy

################################################################################

data = hepfile.initialize()

hepfile.create_group(data,'jet',counter='njet')
hepfile.create_dataset(data,['e','px','py','pz','btag'],group='jet',dtype=float)

hepfile.create_group(data,'muon',counter='nmuon')
hepfile.create_dataset(data,['e','px','py','pz','q'],group='muon',dtype=float)

hepfile.create_group(data,'electron',counter='nelectron')
hepfile.create_dataset(data,['e','px','py','pz','q'],group='electron',dtype=float)

hepfile.create_group(data,'photon',counter='nphoton')
hepfile.create_dataset(data,['e','px','py','pz'],group='photon',dtype=float)

hepfile.create_group(data,'MET',counter='nMET')
hepfile.create_dataset(data,['pt','phi'],group='MET',dtype=float)

event = hepfile.create_single_event(data)

nentries = 10000

#print(data)
#print(event)

#'''
for i in range(0,nentries):

    if i%1000==0:
        print(i)

    hepfile.clear_event(event)

    njet = np.random.randint(10)
    event['jet/njet'] = njet
    for n in range(njet):
        px = 300*beta(2,9)
        py = 300*beta(2,9)
        pz = 300*beta(2,9)
        mass = 5*beta(2,9)
        energy = calc_energy(mass,px,py,pz)
        event['jet/px'].append(px)
        event['jet/py'].append(py)
        event['jet/pz'].append(pz)
        event['jet/e'].append(energy)
        event['jet/btag'].append(np.random.random())

    nmuon = np.random.randint(10)
    event['muon/nmuon'] = nmuon
    for n in range(nmuon):
        px = 300*beta(2,9)
        py = 300*beta(2,9)
        pz = 300*beta(2,9)
        mass = 0.105
        energy = calc_energy(mass,px,py,pz)
        event['muon/px'].append(px)
        event['muon/py'].append(py)
        event['muon/pz'].append(pz)
        event['muon/e'].append(energy)
        event['muon/q'].append(2*np.random.randint(2) - 1)

    nelectron = np.random.randint(10)
    event['electron/nelectron'] = nelectron
    for n in range(nelectron):
        px = 300*beta(2,9)
        py = 300*beta(2,9)
        pz = 300*beta(2,9)
        mass = 0.000511
        energy = calc_energy(mass,px,py,pz)
        event['electron/px'].append(px)
        event['electron/py'].append(py)
        event['electron/pz'].append(pz)
        event['electron/e'].append(energy)
        event['electron/q'].append(2*np.random.randint(2) - 1)

    nphoton = np.random.randint(10)
    event['photon/nphoton'] = nphoton
    for n in range(nphoton):
        px = 300*beta(2,9)
        py = 300*beta(2,9)
        pz = 300*beta(2,9)
        mass = 0.0
        energy = calc_energy(mass,px,py,pz)
        event['photon/px'].append(px)
        event['photon/py'].append(py)
        event['photon/pz'].append(pz)
        event['photon/e'].append(energy)
        
    hepfile.pack(data,event)

print("Writing the file...")
#hdfile = write_to_file('output.hdf5',data)
hdfile = hepfile.write_to_file('HEP_random_file_LARGE.hdf5',data,comp_type='gzip',comp_opts=9)
#'''




import numpy as np
import sys
sys.path.append('../hepfile')
#from write import *
import hepfile

data = hepfile.initialize()

hepfile.create_group(data,'jet',counter='njet')
hepfile.create_dataset(data,['e','px','py','pz'],group='jet',dtype=float)

hepfile.create_group(data,'muons',counter='nmuon')
hepfile.create_dataset(data,['e','px','py','pz'],group='muons',dtype=float)

event = hepfile.create_single_event(data)

#'''
for i in range(0,1000):

    hepfile.clear_event(event)

    njet = 5
    event['jet/njet'] = njet

    for n in range(njet):
        event['jet/e'].append(np.random.random())
        event['jet/px'].append(np.random.random())
        event['jet/py'].append(np.random.random())
        event['jet/pz'].append(np.random.random())

    hepfile.pack(data,event)

print("Writing the file...")
#hdfile = hepfile.write_to_file('output.hdf5',data)
hdfile = hepfile.write_to_file('output.hdf5',data,comp_type='gzip',comp_opts=9)
#'''




import numpy as np
import h5py as h5
import datetime
import sys
import hepfile


################################################################################
def initialize():

    """ Creates an empty data dictionary

    Returns:

	**data** (dict): An empty data dictionary

    """

    data = {}
    data["groups"] = {}
    data["datasets_and_counters"] = {}
    data["list_of_counters"] = []

    # For singleton entries, variables with only one entry per event.
    data["groups"]["_SINGLETON_"] = ["INDEX"]
    data["datasets_and_counters"]["_SINGLETON_"] = "_SINGLETON_/INDEX"
    data["list_of_counters"].append("_SINGLETON_/INDEX")
    data["_SINGLETON_/INDEX"] = []
    return data


################################################################################
def clear_event(data):

    """ Clears the data from the data dictionary - should the name of the function change?

    Args:
	**data** (dict): The dictionary to be cleared

    """

    for key in data.keys():
        if type(data[key]) == list:
            data[key].clear()

        #'''
        # Is this the right thing to do here?????
        elif type(data[key]) == int:
            data[key] = 0
        elif type(data[key]) == float:
            data[key] = 0.0
        #'''


################################################################################
# Create a single event (dictionary) that will eventually be used to fill
# the overall dataset
################################################################################
def create_single_event(data):

    """ Creates an event dictionary that will be used to collect data and then
    packed into the the master data dictionary.

    Args:
	**data** (dict): Data dictionary that will hold all the data from the events.

    Returns:
	**event** (dict): The new event dictionary with keys and no event information

    """

    event = {}

    for k in data.keys():
        if k[-5:] == "index":
            event[k] = data[k]
        elif k in data["groups"]["_SINGLETON_"]:
            event[k] = None
        elif k in data["list_of_counters"]:
            event[k] = 0
        else:
            event[k] = data[k].copy()

    return event


################################################################################
# This adds a group in the dictionary, similar to
# a la CreateBranch in ROOT
################################################################################
def create_group(data, groupname, counter=None):

    """ Adds a group in the dictionary

    Args:
	**data** (dict): Dictionary to which the group will be added

	**groupname** (string): Name of the group to be added

	**counter** (string): Name of the counter key. None by default

    """

    keys = data.keys()

    # Put the counter in the dictionary first.
    """
    if counter is not None:
        data['datasets_and_counters'][groupname] = counter
        keyfound = False
        for k in keys:
            if counter == k:
                keyfound = True
        if keyfound == False:
            data[counter] = []
    """

    # Then put the group and any datasets in there next.
    keyfound = False
    for k in keys:
        if groupname == k:
            print("\033[1m%s\033[0m is already in the dictionary!" % (groupname))
            keyfound = True
            break
    if keyfound == False:
        # data[groupname] = []
        data["groups"][groupname] = []
        print("Adding group \033[1m%s\033[0m" % (groupname))
        if counter is not None:
            data["groups"][groupname].append(counter)
            name = "%s/%s" % (groupname, counter)
            # data['datasets_and_counters'][groupname] = counter
            data["datasets_and_counters"][groupname] = name
            if name not in data["list_of_counters"]:
                data["list_of_counters"].append(name)
            data[name] = []
            print(
                "Adding a counter for \033[1m%s\033[0m as \033[1m%s\033[0m"
                % (groupname, counter)
            )
        else:
            print("----------------------------------------------------")
            print("There is no counter to go with group \033[1m%s\033[0m" % (groupname))
            print("Are you sure that's what you want?")
            print("-----------------------------------------------------")


################################################################################
# This adds a group in the dictionary, similar to
# a la CreateBranch in ROOT
################################################################################
def create_dataset(data, datasets, group=None, dtype=None):

    """ Adds a dataset to a group in a dictionary. If the group does not exist, it will be created.

    Args:
	**data** (dict): Dictionary that contains the group
	
	**datasets** (list): Dataset to be added to the group (This doesn't have to be a list)

	**group** (string): Name of group the dataset will be added to.  None by default

	**dtype** (type): The data type. None by default - I don't think this is every used 

    Returns:
	**-1**: If the group is None


    """

    keys = data.keys()

    if group is None:
        print("-----------------------------------------------")
        print("You need to assign this dataset(s) to a group!")
        print("Groups are not added")
        print("-----------------------------------------------")

        if type(datasets) != list:
            datasets = [datasets]

        for dataset in datasets:
            keyfound = False
            for k in data["groups"]["_SINGLETON_"]:
                if dataset == k:
                    print("\033[1m%s\033[0m is already in the dictionary!" % (dataset))
                    keyfound = True
            if keyfound == False:
                print(
                    "Adding dataset \033[1m%s\033[0m to the dictionary as a SINGLETON."
                    % (dataset)
                )
                data["groups"]["_SINGLETON_"].append(dataset)
                data[dataset] = []
                # counter_name = "%s/%s" % (group,counter)
                data["datasets_and_counters"][dataset] = "_SINGLETON_/INDEX"

        return 0

    # Put the counter in the dictionary first.
    keyfound = False
    for k in data["groups"]:
        if group == k:
            keyfound = True

    # NEED TO FIX THIS PART SO THAT IT FINDS THE RIGHT COUNTER FROM THE GROUP
    if keyfound == False:
        print("Your group, \033[1m%s\033[0m is not in the dictionary yet!" % (group))
        counter = "n%s" % (group)
        print("Adding it, along with a counter of \033[1m%s\033[0m" % (counter))
        create_group(data, group, counter=counter)

    # Then put the datasets into the group in there next.
    if type(datasets) != list:
        datasets = [datasets]

    for dataset in datasets:
        keyfound = False
        name = "%s/%s" % (group, dataset)
        for k in keys:
            if name == k:
                print("\033[1m%s\033[0m is already in the dictionary!" % (name))
                keyfound = True
        if keyfound == False:
            print(
                "Adding dataset \033[1m%s\033[0m to the dictionary under group \033[1m%s\033[0m."
                % (dataset, group)
            )
            data[name] = []
            data["groups"][group].append(dataset)

            # Add a counter for this dataset for the group with which it is associated.
            counter = data["datasets_and_counters"][group]
            # counter_name = "%s/%s" % (group,counter)
            data["datasets_and_counters"][name] = counter

    return 0


################################################################################
def pack(data, event):

    """ Takes the data from an event and packs it into the data dictionary, 
    intelligently, so that it can be stored and extracted efficiently. 
    (This is analagous to the ROOT TTree::Fill() member function.

    Args:
	**data** (dict): Data dictionary to hold the entire dataset.

	**event** (dict): Event to be packed into data.

    """

    keys = list(event.keys())

    for key in keys:
        # print(key)

        if (
            key == "datasets_and_counters"
            or key == "groups"
            or key == "list_of_counters"
        ):
            continue

        # The singletons will only have 1 entry per event
        if key == "_SINGLETON_/INDEX":
            data[key].append(1)
            continue

        # if key[-5:] == 'counter':
        # continue
        if type(event[key]) == list:
            value = event[key]
            if len(value) > 0:
                data[key] += value
            """
            else:
                # No entries for this event
                #print(key)
                counter = data['datasets_and_counters'][key]
                data[counter].append(0)
                if counter in keys:
                    keys.remove(counter)
            """
        else:
            # This is for counters and SINGLETONS
            if key in data["groups"]["_SINGLETON_"]:
                if event[key] == None:
                    print(
                        "\n\033[1m%s\033[0m is part of the SINGLETON group and is expected to have a value for each event."
                        % (key)
                    )
                    print("However it is None...exiting.\n")
                    exit()
            else:
                data[key].append(event[key])


################################################################################
def convert_list_and_key_to_string_data(datalist, key):

    """ Converts data dictionary to a string

    Args:
	**datalist** (list): A list to be saved as a string.

    Returns:
	**key** (string): We will assume that this will be unpacked as a dictionary,
                      and this will be the key for the list in that dictionary.

    """

    a = np.string_(key)

    mydataset = []
    b = np.string_("")
    nvals = len(datalist)
    for i, val in enumerate(datalist):
        # print(val,type(val))
        b += np.string_(val)
        if i < nvals - 1:
            b += np.string_("__:__")
    mydataset.append([a, b])

    return mydataset


################################################################################

################################################################################
def convert_dict_to_string_data(dictionary):

    """ Converts data dictionary to a string

    Args:
	**dictionary** (dict): Dictionary to be converted to a string

    Returns:
	**mydataset** (string): String representation of the dataset

    """

    keys = dictionary.keys()

    nkeys = len(keys)

    mydataset = []
    for i, key in enumerate(keys):
        # print(i,key,dictionary[key])
        a = np.string_(key)
        b = np.string_(dictionary[key])
        mydataset.append([a, b])

    return mydataset

################################################################################
# This function writes default attributes and metadata to a file.
################################################################################
def write_file_metadata(
    filename, mydict = {}, write_default_values=True, append=True
):
    """ Writes file metadata in the attributes of an HDF5 file

    Args:
    **filename** (string): Name of output file

    **mydict** (dictionary): Metadata desired by user

    **write_default_values** (boolean): True if user wants to write/update the 
                                        default metadata: date, hepfile version, 
                                        h5py version, numpy version, and Python 
                                        version, false if otherwise.

    **append** (boolean): True if user wants to keep older metadata, false otherwise.

    Returns:
    **hdoutfile** (HDF5): File with new metadata

    """

    hdoutfile = h5.File(filename, "a")
    
    non_metadata = ["nentries"]

    if not append:
        for key in hdoutfile.attr.keys():
            if key not in non_metadata:
                del hdoutfile.attrs[key]
    
    if write_default_values:
        hdoutfile.attrs["date"] = datetime.datetime.now().isoformat(sep=' ')
        hdoutfile.attrs["hepfile_version"] = hepfile.__version__
        hdoutfile.attrs["numpy_version"] = np.__version__
        hdoutfile.attrs["h5py_version"] = h5.__version__
        hdoutfile.attrs["python_version"] = sys.version

    for key in mydict:
        hdoutfile.attrs[key] = mydict[key]

    hdoutfile.close()
    print("Metadata added")
    return hdoutfile


################################################################################

################################################################################
def write_to_file(
    filename, data, comp_type=None, comp_opts=None, force_single_precision=True
):

    """ Writes the selected data to an HDF5 file

    Args:
	**filename** (string): Name of output file

	**data** (dictionary): Data to be written into output file

	**comp_type** (string): Type of compression

	**force_single_precision** (boolean): True if data should be written in single precision

    Returns:
	**hdoutfile** (HDF5): File to which the data has been written 

    """

    hdoutfile = h5.File(filename, "w")

    groups = data["groups"].keys()

    # Convert this to a 2xN array for writing to the hdf5 file.
    # This gives us one small list of informtion if we need to pull out
    # small chunks of data
    mydataset = convert_dict_to_string_data(data["datasets_and_counters"])
    dset = hdoutfile.create_dataset(
        "datasets_and_counters",
        data=mydataset,
        dtype="S256",
        compression=comp_type,
        compression_opts=comp_opts,
    )

    # Convert this to a 2xN array for writing to the hdf5 file.
    # This has the groups and the datasets in them.
    mydataset = convert_list_and_key_to_string_data(
        data["groups"]["_SINGLETON_"], "_SINGLETONGROUP_"
    )
    dset = hdoutfile.create_dataset(
        "_SINGLETONGROUP_",
        data=mydataset,
        dtype="S256",
        compression=comp_type,
        compression_opts=comp_opts,
    )

    for group in groups:

        # print(group)

        hdoutfile.create_group(group)
        hdoutfile[group].attrs["counter"] = np.string_(
            data["datasets_and_counters"][group]
        )

        datasets = data["groups"][group]

        # print(datasets)

        for dataset in datasets:

            name = None
            if group == "_SINGLETON_" and dataset is not "INDEX":
                name = dataset
            else:
                name = "%s/%s" % (group, dataset)

            x = data[name]
            if type(x) == list:
                x = np.array(x)

            # Do single precision only, unless specified
            if force_single_precision == True:
                if x.dtype == np.float64:
                    x = x.astype(np.float32)

            hdoutfile.create_dataset(
                name, data=x, compression=comp_type, compression_opts=comp_opts
            )

    # Get the number of events
    counters = data["list_of_counters"]
    nentries = -1
    prevcounter = None
    for i, countername in enumerate(counters):
        ncounter = len(data[countername])
        print("%-32s has %-12d entries" % (countername, ncounter))
        if i > 0 and ncounter != nentries:
            print("-------- WARNING -----------")
            print(
                "%s and %s have differing numbers of entries!"
                % (countername, prevcounter)
            )
            print("-------- WARNING -----------")
            # SHOULD WE EXIT ON THIS?

        if nentries < ncounter:
            nentries = ncounter

        prevcounter = countername


    hdoutfile.attrs["nentries"] = nentries
    hdoutfile.close()

    write_file_metadata(filename)

    return hdoutfile



import h5py as h5
import numpy as np
import matplotlib.pylab as plt
import time

import hepfile

import sys

filename = sys.argv[1]

#data,event = hepfile.load(filename,subset=(0,100000))
data,event = hepfile.load(filename,verbose=False)#,subset=10000)
#data,event = hepfile.load(filename,desired_datasets=['jet','muon'])
#data,event = hepfile.load(filename,desired_datasets=['jet'])
#data,event = hepfile.load(filename,desired_datasets=['jet','muon'],subset=(0,100000))

#print(data['list_of_counters'])

nentries = data['nentries']
print("nentries: ",nentries)

#print(type(data),type(event))

energies = []

#print("A")
#print(data.keys())
#print(data)
#
#print("B")
#print(event.keys())
#print(event)
#x = data['jet/e']

# Print out what has been read in from the files.
'''
for key in event.keys():
    print(key)
'''

for i in range(0,nentries):

    if i%10000==0:
        print(i)

    hepfile.unpack(event,data,n=i)

    #print(event['jet/njet'])

    energy = event['jet/e']

    '''
    for e in energy:
        energies.append(e)
    '''
    energies += energy.tolist()


print(len(energies))

plt.figure()
plt.hist(energies,bins=100,range=(0,500))

#plt.show()



import numpy as np
import sys
import hepfile

#sys.path.append("../hepfile")


def write_file_for_unit_tests():
    data = hepfile.initialize()

    hepfile.create_group(data, "jet", counter="njet")
    hepfile.create_dataset(data, ["e", "px", "py", "pz"], group="jet", dtype=float)

    hepfile.create_group(data, "muons", counter="nmuon")
    hepfile.create_dataset(data, ["e", "px", "py", "pz"], group="muons", dtype=float)

    event = hepfile.create_single_event(data)

    #'''
    for i in range(0, 10):

        hepfile.clear_event(event)

        njet = 5
        event["jet/njet"] = njet

        for n in range(njet):
            event["jet/e"].append(np.random.random())
            event["jet/px"].append(np.random.random())
            event["jet/py"].append(np.random.random())
            event["jet/pz"].append(np.random.random())

        hepfile.pack(data, event)

    print("Writing the file...")
    # hdfile = write_to_file('output.hdf5',data)
    hdfile = hepfile.write_to_file("FOR_TESTS.hdf5", data, comp_type="gzip", comp_opts=9)
    #'''


if __name__ == "__main__":
    write_file_for_unit_tests()



import numpy as np
import h5py as h5
import hepfile 

import time

#import sys
#sys.path.append('./scripts')
#from write_h5hep_file_for_unit_tests import write_h5hep_file_for_unit_tests

from write_file_for_unit_tests import write_file_for_unit_tests

def isEmpty(dictionary):
    test = True
    print(dictionary.keys())
    for key in dictionary.keys():
        print(key)
        print(dictionary[key])
        print(type(dictionary[key]))
        if dictionary[key] is None:
            test = True
        elif type(dictionary[key])==list or type(dictionary[key])==np.ndarray:
            if len(dictionary[key]) > 0:
                test = False

    return test


def test_load():

    write_file_for_unit_tests()

    # This assumes you run nosetests from the h5hep directory and not 
    # the tests directory.
    filename = "FOR_TESTS.hdf5"
    desired_datasets = ['jet','muon']
    subset = 1000

    test_data,test_event = hepfile.load(filename, False, desired_datasets, subset)

    assert isinstance(test_data, dict)
    assert isinstance(test_event, dict)

    assert isEmpty(test_event) == True
    assert isEmpty(test_data) == False

def test_unpack():
	
    # This assumes you run nosetests from the h5hep directory and not 
    # the tests directory.
    filename = "FOR_TESTS.hdf5"
    desired_datasets = ['jet','muon']
    subset = 1000

    event, data = hepfile.load(filename, False, desired_datasets, subset)

    hepfile.unpack(data, event)

    assert isEmpty(event) == False


def test_get_nentries():
	
    # This assumes you run nosetests from the h5hep directory and not 
    # the tests directory.
    filename = "FOR_TESTS.hdf5"

    nentries = hepfile.get_nentries(filename)

    assert nentries == 10

def test_get_file_metadata():

    filename = "FOR_TESTS.hdf5"

    metadata = hepfile.get_file_metadata(filename)

    assert 'date' in metadata
    assert 'hepfile_version' in metadata
    assert 'h5py_version' in metadata
    assert 'numpy_version' in metadata
    assert 'python_version' in metadata

    #Check default attributes are strings
    assert isinstance(metadata['date'],str)
    assert isinstance(metadata['hepfile_version'],str)
    assert isinstance(metadata['h5py_version'],str)
    assert isinstance(metadata['numpy_version'],str)
    assert isinstance(metadata['python_version'],str)








import numpy as np
import h5py as h5
import hepfile

import matplotlib.pylab as plt
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import colors

# Number of layers
nlayers = 3

# Number of events
nevents = 4

# Assign dimensions
z_size = [20, 24, 28]

# Max number of pixels in each layer
# Assume i pixels per ten degrees
maxpix = []
for i in range(0,nlayers):
    #maxpix.append(z_size[i]*(i+1)*36)
    maxpix.append(z_size[i]*(i+1)*18)

# Initialize the data dictionary for a single matrix
data = hepfile.initialize()

#Create groups
hepfile.create_group(data, "pixels", counter="npixels")

# Create datasets within the groups
hepfile.create_dataset(data, ["z","theta","r","value"], group="pixels")

# Create an event from the data container
event = hepfile.create_single_event(data)

for i in range(0,nevents):

    npixtot = 0
    ## Random selection of number of hit pixels in the layer
    for j in range(nlayers): 
        npix = np.random.randint(0,maxpix[j])

        # For each point, randomly select a z and theta
        # Repeats are ok in this example
        for i in range(0,npix):
            event['pixels/z'].append(np.random.randint(0,z_size[j]) - z_size[j]/2)
            event['pixels/r'].append(j)
            event['pixels/theta'].append(np.random.random()*2*np.pi)
            event['pixels/value'].append(np.random.randint(1,70))

        # Keep a count of the total number of pixels
        npixtot += npix

    # Counters are the same here because there is always one data point per point
    # They could be different if there was more data collected for each point
    event['pixels/npixels'] = npixtot

    # Add the event to the matrix dictionary
    hepfile.pack(data,event)

    # Clear the event matrix for the next iteration
    hepfile.clear_event(event)


###############
# Write to file
###############
filename = 'cylindrical_example_output.hdf5'
hdfile = hepfile.write_to_file(filename,data,comp_type='gzip',comp_opts=9)

################
# Read from file
################
data,event = hepfile.load(filename)
nevents = data['nentries']
#print(data)
#print("nevents: ",nevents)

fig = plt.figure(figsize=(10,10))

for i in range(0,nevents):
    
    hepfile.unpack(event,data,n=i)

    npixels = event['pixels/npixels']
    #npixels = data['pixels/npixels'][i]
    z = event['pixels/z']
    r = event['pixels/r']
    theta = event['pixels/theta']
    d = event['pixels/value']

    x = (r+1)*np.cos(theta)
    y = (r+1)*np.sin(theta)
    
    #x1 = np.linspace(-1,1,100)
    #y1 = np.linspace(-1,1,100)
    #z1 = np.linspace(0,10,100)

    #zipped = zip(x,y,z,d)
    #sorted(zipped, key=lambda g: g[2])

    ax = fig.add_subplot(2, 2, i+1, projection='3d')
    ax.scatter(x,y,z,c=d, alpha=0.2)
    #ax.scatter(x,y,z,c=r, alpha=0.2)
    #ax.plot_surface(x1,y1,z1)
    #ax.plot_surface(x1,y1,z1)
    #ax.set_zlim(0, 10)

plt.show()





import h5py as h5
import numpy as np

################################################################################
def load(filename=None, verbose=False, desired_datasets=None, subset=None):

    """ Reads all, or a subset of the data, from the HDF5 file to fill a data dictionary.
    Returns an empty dictionary to be filled later with select events.

    Args:
	**filename** (string): Name of the input file
	
	**verbose** (boolean): True if debug output is required

	**desired_datasets** (list): Datasets to be read from input file

	**subset** (int): Number of events to be read from input file

    Returns:
	**ourdata (dict): Selected data from HDF5 file
	
	**event** (dict): An empty event dictionary to be filled by individual events

    """

    f = None
    if filename != None:
        f = h5.File(filename, "r+")
    else:
        print("No filename passed in! Can't open file.\n")
        return None

    ourdata = {}
    ourdata["datasets_and_counters"] = {}
    ourdata["datasets_and_indices"] = {}
    ourdata["list_of_counters"] = []
    ourdata["all_datasets"] = []

    ourdata["nentries"] = f.attrs["nentries"]
    if subset is not None:
        if type(subset) == int:
            subset = (0, subset)
        ourdata["nentries"] = subset[1] - subset[0]

    event = {}

    # Get the datasets and counters
    dc = f["datasets_and_counters"]
    for vals in dc:
        # The decode is there because vals were stored as numpy.bytes
        counter = vals[1].decode()
        index = "%s_INDEX" % (counter)
        ourdata["datasets_and_counters"][vals[0].decode()] = counter
        ourdata["datasets_and_indices"][vals[0].decode()] = index
        ourdata["list_of_counters"].append(vals[1].decode())
        ourdata["all_datasets"].append(vals[0].decode())
        ourdata["all_datasets"].append(vals[1].decode())  # Get the counters as well

    # We may have added some strings (like counters) multiple times.
    ourdata["list_of_counters"] = np.unique(ourdata["list_of_counters"]).tolist()
    ourdata["all_datasets"] = np.unique(ourdata["all_datasets"]).tolist()

    # Pull out the SINGLETON datasets
    sg = f["_SINGLETONGROUP_"][0]  # This is a numpy array of strings
    decoded_string = sg[1].decode()

    vals = decoded_string.split("__:__")
    vals.remove("INDEX")

    ourdata["_SINGLETON_"] = vals

    # Get the list of datasets and groups, but remove the
    # 'datasets_and_counters', as that is a protected key.
    entries = ourdata["all_datasets"]

    ########################################################
    # Only keep select data from file
    ########################################################
    if desired_datasets is not None:
        if type(desired_datasets) != list:
            desired_datasets = list(desired_datasets)

        # Count backwards because we'll be removing stuff as we go.
        i = len(entries) - 1
        while i >= 0:
            entry = entries[i]

            is_dropped = True
            for desdat in desired_datasets:
                if desdat in entry:
                    is_dropped = False
                    break

            if is_dropped == True:
                print("Not reading out %s from the file...." % (entry))
                entries.remove(entry)

            i -= 1
    #######################################################

    if verbose == True:
        print("Datasets and counters:")
        print(ourdata["datasets_and_counters"])
        print("\nDatasets and indices:")
        print(ourdata["list_of_counters"])

    # Pull out the counters first and build the indices
    print("Building the indices...")
    for name in ourdata["list_of_counters"]:
        if subset is not None:
            ourdata[name] = f[name][subset[0] : subset[1]]
        else:
            ourdata[name] = f[name][:]

        # counter = f[name].value
        indexname = "%s_INDEX" % (name)
        index = np.zeros(len(ourdata[name]), dtype=int)
        start = 0
        nentries = len(index)
        for i in range(0, nentries):
            index[i] = start
            nobjs = ourdata[name][i]
            start = index[i] + nobjs
        ourdata[indexname] = index
    print("Built the indices!")

    # Loop over the entries we want and pull out the data.
    for name in entries:

        # The decode is there because counter is a numpy.bytes object
        counter = None
        if name not in ourdata["list_of_counters"]:
            counter = ourdata["datasets_and_counters"][name]

        if verbose == True:
            print(f[name])

        data = f[name]
        # for data in f[name]:
        if type(data) == h5.Dataset:
            datasetname = name

            if subset is not None:
                ourdata[datasetname] = data[subset[0] : subset[1]]
            else:
                ourdata[datasetname] = data[:]

            event[datasetname] = None  # This will be filled for individual events
            if verbose == True:
                print(data)

    f.close()
    print("Data is read in and input file is closed.")

    return ourdata, event


################################################################################


################################################################################
def unpack(event, data, n=0):

    """ Fills the event dictionary with selected events.

    Args:

	**event** (dict): Event dictionary to be filled

	**data** (dict): Data dictionary used to fill the event dictionary

    """

    keys = event.keys()

    for key in keys:

        # if "num" in key:
        # IS THERE A WAY THAT THIS COULD BE FASTER?
        # print(data['list_of_counters'],key)
        if key in data["list_of_counters"] or key in data["_SINGLETON_"]:
            # print("here! ",key)
            event[key] = data[key][n]

        elif "INDEX" not in key:  # and 'Jets' in key:
            indexkey = data["datasets_and_indices"][key]
            numkey = data["datasets_and_counters"][key]

            if len(data[indexkey]) > 0:
                index = data[indexkey][n]

            if len(data[numkey]) > 0:
                nobjs = data[numkey][n]
                event[key] = data[key][index : index + nobjs]


################################################################################
def get_nentries(filename):

    """ Get the number of entries in the file.

    """

    f = h5.File(filename, "r+")

    a = f.attrs

    if a.__contains__("nentries"):
        nentries = a.get("nentries")
        f.close()
        return nentries
    else:
        print('\nFile does not contain the attribute, "nentries"\n')
        f.close()
        return None


################################################################################
def get_file_metadata(filename):

    """ Get the file metadata and return it as a dictionary

    """

    f = h5.File(filename, "r+")

    a = f.attrs

    if len(a)<1:
        print(f"No metadata in file {filename}!")
        print(f"File has no attributes.\n")
        f.close()
        return None

    metadata = {}

    for key in a.keys():
        metadata[key] = a[key]
    
    f.close()

    return metadata


################################################################################
 
################################################################################
def print_file_metadata(filename):

    """ Pretty print the file metadata 

    """

    metadata = get_file_metadata(filename)

    if metadata is None:
        return None


    output = ""

    keys = list(metadata.keys())

    first_keys_to_print = ['date','nentries']

    keys_already_printed = []

    # Print the basics first
    for key in first_keys_to_print:
        if key in keys:
            val = metadata[key]
            output += f"{key:<20s} : {val}\n"
            keys_already_printed.append(key)

    # Print the versions next
    for key in keys:
        if key in keys_already_printed:
            continue

        if key.find('version')>=0:
            val = metadata[key]
            output += f"{key:<20s} : {val}\n"
            keys_already_printed.append(key)

    # Print the read of the metadata
    for key in keys:
        if key in keys_already_printed:
            continue

        val = metadata[key]
        output += f"{key:<20s} : {val}\n"
        keys_already_printed.append(key)

    print(output)

    return output



################################################################################
 



import numpy as np
import h5py as h5
import hepfile 

import time

import sys
sys.path.append('./scripts')
#from write_h5hep_file_for_unit_tests import write_h5hep_file_for_unit_tests

def isEmpty(dictionary):
    test = True
    print(dictionary.keys())
    for key in dictionary.keys():
        print(key)
        print(dictionary[key])
        print(type(dictionary[key]))
        if dictionary[key] is None:
            test = True
        elif type(dictionary[key])==list or type(dictionary[key])==np.ndarray:
            if len(dictionary[key]) > 0:
                test = False

    return test

def test_initialize():

    test_data = hepfile.initialize()

    assert isinstance(test_data, dict)
    assert test_data['groups']['_SINGLETON_'] ==  ['INDEX']
    assert test_data['datasets_and_counters']['_SINGLETON_'] == '_SINGLETON_/INDEX'
    assert test_data['list_of_counters'] == ['_SINGLETON_/INDEX']
    assert test_data['_SINGLETON_/INDEX'] == []


def test_clear_event():
	
    # This assumes you run nosetests from the h5hep directory and not 
    # the tests directory.
    filename = "FOR_TESTS.hdf5"
    desired_datasets = ['jet','muon']
    subset = 1000

    data, event = hepfile.load(filename, False, desired_datasets, subset)

    hepfile.clear_event(event)

    assert isEmpty(event) == True

def test_create_single_event():

    data = hepfile.initialize()

    hepfile.create_group(data,'jet',counter='njet')
    hepfile.create_dataset(data,['e','px','py','pz'],group='jet',dtype=float)

    hepfile.create_group(data,'muons',counter='nmuon')
    hepfile.create_dataset(data,['e','px','py','pz'],group='muons',dtype=float)

    test_event = hepfile.create_single_event(data)

    assert isEmpty(test_event) == False
    assert isinstance(test_event, dict)

def test_create_group():

    data = hepfile.initialize()
    hepfile.create_group(data,'jet',counter='njet')

    assert isEmpty(data['groups']) == False
    assert 'jet/njet' in data.keys()


def test_create_dataset():

    data = hepfile.initialize()
    hepfile.create_group(data,'jet',counter='njet')
    hepfile.create_dataset(data,['e','px','py','pz'],group='jet',dtype=float)


    assert isEmpty(data['groups']) == False
    assert 'jet/njet' in data.keys()
    assert 'jet/e' in data.keys()
    assert 'jet/px' in data.keys()
    assert 'jet/e' in data['datasets_and_counters'].keys()
    assert data['datasets_and_counters']['jet/e'] == 'jet/njet'


def test_write_file_metadata():

    filename = "FOR_TESTS.hdf5"
    file = h5.File(filename, "r")

    #Check default attribute existence
    assert 'date' in file.attrs.keys()
    assert 'hepfile_version' in file.attrs.keys()
    assert 'h5py_version' in file.attrs.keys()
    assert 'numpy_version' in file.attrs.keys()
    assert 'python_version' in file.attrs.keys()

    #Check default attributes are strings
    assert isinstance(file.attrs['date'],str)
    assert isinstance(file.attrs['hepfile_version'],str)
    assert isinstance(file.attrs['h5py_version'],str)
    assert isinstance(file.attrs['numpy_version'],str)
    assert isinstance(file.attrs['python_version'],str)

    file.close()

    #Adding a new attribute
    hepfile.write_file_metadata(filename, {'author':'John Doe'})
    file = h5.File(filename, "r")

    assert 'author' in file.attrs.keys()
    assert file.attrs['author'] == 'John Doe'

    file.close()
    








